#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Some functions implemented in module ``types`` in Python 3 but not in
Python 2 needed for '_meta*' implementation.

'''
__all__ = ['new_class', 'prepare_class', 'get_exec_body']


from types import new_class, prepare_class  # noqa
from types import MappingProxyType  # noqa


def get_exec_body(**kwargs):
    '''Return an `exec_body` function that update `ns` with `kwargs`.'''
    def exec_body(ns):
        ns.update(kwargs)
        return ns
    return exec_body



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Python 2 and Python 3 compatibility.

The name comes from (Manu's idea') "2 raised to the power of 3".

There is an existing library written by "Benjamin Peterson" named six_, both
(`xoutil.eight` and `six`) can be used together since this module don't claim
to be a replacement of six_, just some extra extensions.  Nevertheless, there
are some simple definitions that even when are in six_ also are defined here.

This package also fixes some issues from PyPy interpreter.

.. _six: https://pypi.python.org/pypi/six

'''
from xoutil.versions import python_version


try:
    from hashlib import sha1 as sha    # noqa
except ImportError:
    from sha import sha    # noqa

base_string = str
string_types = (str, )
integer_types = int,
long_int = int
class_types = type,
text_type = str
unichr = chr

binary_type = bytes
UnicodeType = text_type
StringTypes = string_types

#: Define a tuple with both base types in Python 2 and containing only `type`
#: in Python 3.
ClassTypes = class_types


# The `memoryview`:class: API is similar but not exactly the same as that
# of `buffer`.
buffer = memoryview


# Python versions

_pyver = python_version.to_float()
_py2 = python_version == 2
_py3 = python_version == 3
_py33 = python_version >= 3.3
_py34 = python_version >= 3.4
_pypy = python_version.pypy


typeof = type


def force_type(obj):
    '''Ensure return a valid type from `obj`.

    If `obj` is already a "type", return itself, else obtain its type.

    '''
    return obj if isinstance(obj, class_types) else typeof(obj)


def type_name(obj):
    '''Return the type name.'''
    return typeof(obj).__name__


from sys import intern  # noqa

input = input
range = range
ascii = ascii    # noqa
hex, oct, filter, map, zip = hex, oct, filter, map, zip


def iterkeys(d):
    '''Return an iterator over the keys of a dictionary.'''
    return iter((d.keys if _py3 else d.iterkeys)())


def itervalues(d):
    '''Return an iterator over the values of a dictionary.'''
    return iter((d.values if _py3 else d.itervalues)())


def iteritems(d):
    '''Return an iterator over the (key, value) pairs of a dictionary.'''
    return iter((d.items if _py3 else d.iteritems)())


from io import StringIO  # noqa

import builtins    # noqa
__builtin__ = builtins


callable = getattr(__builtin__, 'callable')    # noqa
exec_ = getattr(builtins, 'exec')    # noqa


def execfile(filename, globals=None, locals=None):
    """Read and execute a Python script from a file.

    The globals and locals are dictionaries, defaulting to the current
    globals and locals.  If only globals is given, locals defaults to it.

    """
    import sys
    if globals is None:
        frame = sys._getframe(1)
        globals = frame.f_globals
        if locals is None:
            locals = frame.f_locals
        del frame
    elif locals is None:
        locals = globals
    with open(filename, "r") as f:
        source = f.read()
        code = compile(source, filename, 'exec')
        return exec_(code, globals, locals)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
'''Functional tools for functions that returns iterators (generators, etc.)

.. warning:: This module is experimental.  It may be removed completely, moved
   or otherwise changed.

'''
from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from functools import reduce


def iter_compose(*fs):
    '''Compose functions that take an item and return an iterator.

    For two functions, ``iter_compose(g, f)`` is roughly equivalent to::

       lambda x: (z for y in f(x) for z in g(y))

    In general this is, ``reduce(_compose, fs, lambda x: [x])``; where
    ``_compose`` is the lambda for two arguments.

    .. versionadded:: 1.9.6

    '''
    def _compose(g, f):
        # (>>.) :: Monad m => (b -> m c) -> (a -> m b) -> a -> m c
        # g >>. f = \x -> f x >>= g
        #
        # In the list monad:
        #
        # g >>. f = \x -> concat (map g (f x))
        return lambda x: (z for y in f(x) for z in g(y))

    if len(fs) == 2:
        # optimize a bit so that we can avoid the 'lambda x: [x]' for common
        # cases.
        return _compose(*fs)
    else:
        return reduce(_compose, fs, lambda x: [x])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Compatibility to Python's `urllib`:mod: standard module.

.. deprecated:: 2.0.0

'''

from urllib import error, parse, request, response, robotparser  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Solve compatibility issues for exceptions handling.

This module tries to solve, as much as possible, differences between how the
`raise``, and ``try ... except`` statements are executed between Python
versions 2 and 3.  See `Exception Chaining and Embedded Tracebacks
<3134>`:pep:.

First issue is that before Python 3, module `exceptions` was defined, but not
anymore.  We decided not to implement this concept in `xoutil.future`:mod:
package because all these exception classes are built-ins in both Python major
versions, so use any of them directly.  But `StandardError`:class: is
undefined in Python 3, we introduce it here.

Raising an exception it's the biggest difference between both major versions,
starting with that there is no way to achieve syntax compatibility.  In
Python 2 the syntax for ``raise`` is::

  "raise" [type ["," value ["," traceback]]]

and in Python 3::

  "raise" [error[.with_traceback(traceback)] ["from" cause]]

There are two partial solutions in this module:

- Using the function `throw`:func: function instead the ``raise`` statement::

    throw([error[, traceback=traceback[, cause=cause]]])

  This is suitable only when you need the trace-back information is fully
  tracked when you raise the exception (see below).

- Using the function `grab`:func: the context_ attributes are assigned, but
  the trace-back information is not tracked in the system call context.

    raise grab(error[, traceback=traceback[, cause=cause]])

.. _context:

Python 3 is consistent by always assigning context information in the
following scenarios:

1. If an exception is raised inside an exception handler or a finally clause:
   the previous exception is then attached as the new exception's
   ``__context__`` attribute.  This is guaranteed in Python 2 using either the
   `grab`:func: function or the `throw`:func: function.

2. Clause ``from`` is used for exception chaining, the expression must be
   another exception instance or class, which will be attached to the newly
   raised exception in the ``__cause__`` attribute.  This information is
   printed when the raised exception is not handled, but in Python 2 this
   information can be custom managed by using either `grab`:func: or
   `throw`:func: function.

3. Trace-back information is automatically attached to the attribute
   ``__traceback__`` when an exception is raised.  You can set your own
   trace-back using the ``with_traceback`` method (or in the compatibility
   function `throw`:func: using ``traceback`` argument).  This can be
   partially guaranteed in Python 2 by using the `catch`:func: function in
   some exception handler or finally clause.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoutil.symbols import Unset

try:
    StandardError    # noqa
except NameError:
    StandardError = Exception

try:
    BaseException    # noqa
except NameError:
    BaseException = StandardError


try:
    with_traceback = BaseException.with_traceback    # only in python 3
    _py3 = True
except AttributeError:
    _py3 = False

    def with_traceback(self, tb):
        '''set self.__traceback__ to `tb` and return self.'''
        from types import TracebackType
        if tb is None or isinstance(tb, TracebackType):
            self.__traceback__ = tb
            return self
        else:
            raise TypeError('__traceback__ must be a traceback or None')


def _force_error(error):
    if isinstance(error, type) and issubclass(error, BaseException):
        error = error()
    return error


def grab(error, traceback=Unset, cause=Unset):
    '''Prepare an error with traceback, and context to raise it.'''
    error = _force_error(error)
    if traceback is not Unset:
        error = with_traceback(error, traceback)
    if cause is not Unset:
        error.__cause__ = _force_error(cause)
    if not _py3 and error is not None:
        from ._errors2 import _fix_context
        _fix_context(error, context=catch())
    return error


def throw(error=Unset, traceback=Unset, cause=Unset):
    '''Unify syntax for raising an error with trace-back information.

    Instead of using the Python ``raise`` statement, use ``throw``::

      "raise" [error[.with_traceback(traceback)] ["from" cause]]

    becomes::

       throw([error, [traceback=traceback, ][cause=cause]])

    If `traceback` argument is not given, it is looked up in the error.


    '''
    if error is Unset:
        if traceback is Unset and cause is Unset:
            import sys
            _, error, traceback = sys.exc_info()
        else:
            raise SyntaxError('invalid syntax')
    if error is None:
        raise RuntimeError('No active exception to re-raise')
    else:
        error = grab(error, traceback=traceback, cause=cause)
        if _py3:
            if cause is None:
                raise error
            else:
                from ._errors3 import raise3
                raise3(error)
        elif error.__traceback__ is None:
            raise error
        else:
            from ._errors2 import raise2
            raise2(error)


def catch(base=Unset):
    '''Fix ``sys.exc_info`` context in Python 2.

    :param base: A subclass of `BaseException`:class: or Unset.

    Next example of Python 3 doesn't work in Python 2::

      >>> def test():
      ...     try:
      ...         try:
      ...             1/0
      ...         except Exception as error:
      ...             raise RuntimeError('xxx') from error
      ...     except Exception as error:
      ...         return error

      >>> error = test()

      >>> error.__context__ is error.__cause__
      True

      >>> isinstance(error.__cause__, ZeroDivisionError)
      True

    In this case, we can solve most issues by two replacements (``throw`` and
    ``catch``)::

      >>> def test():
      ...     try:
      ...         try:
      ...             1/0
      ...         except catch(Exception) as error:
      ...             throw(RuntimeError('xxx'), cause=error)
      ...     except catch(Exception) as error:
      ...         return error

    If base is unset, the catched system error value is returned, so never use
    this function as::

      >>> try:
      ...     1/0
      ... except catch():
      ...     pass

    '''
    if not _py3:
        import sys
        from ._errors2 import _fix_context
        _, error, traceback = sys.exc_info()
        if error is not None:
            _fix_context(error, traceback=traceback)
            if base is Unset:
                base = error
    return base


def catch_traceback(error, traceback):
    '''Catch a trace-back information in a corresponding error.'''
    if not _py3 and error is not None:
        from ._errors2 import _fix_context
        _fix_context(error, traceback=traceback)


def traceof(error):
    '''Get the trace-back information of the given `error`.

    Python 3 is consistent by always assigning the trace-back information in
    the attribute ``__traceback__``, but in Python 2 it's a "little tricky" to
    obtain a good solution, and not always, only in some contexts.

    Returns None if it could not be found.

    '''
    if not _py3:
        from ._errors2 import _catch
        _catch(error)
    return error.__traceback__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import absolute_import as _py3_abs_import

from xoutil.symbols import Unset


def _fix_context(error, **kwds):
    names = ('traceback', 'cause', 'context')
    i, count = 0, len(names)
    while i < count:
        name = '__{}__'.format(names[i])
        value = kwds.get(names[i], Unset)
        if value is not Unset:
            setattr(error, name, value)
            if isinstance(value, BaseException):
                _fix_context(value)
        elif not hasattr(error, name):
            setattr(error, name, None)
        i += 1


def raise2(error):
    raise error, None, error.__traceback__    # noqa



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Implements the metaclass() function using the Py3k syntax.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ._meta import Mixin    # noqa

from ._meta3 import metaclass
metaclass.__doc__ = '''Define the metaclass of a class.

    .. versionadded:: 1.7.0

    .. deprecated:: 2.0.0  Not needed in a world without Python 2.

    This function allows to define the metaclass of a class equally in Python
    2 and 3.

    Usage::

     >>> class Meta(type):
     ...   pass

     >>> class Foobar(metaclass(Meta)):
     ...   pass

     >>> class Spam(metaclass(Meta), dict):
     ...   pass

     >>> type(Spam) is Meta
     True

     >>> Spam.__bases__ == (dict, )
     True

    .. versionadded:: 1.5.5 The `kwargs` keywords arguments with support for
       ``__prepare__``.

    Metaclasses are allowed to have a ``__prepare__`` classmethod to return
    the namespace into which the body of the class should be evaluated.  See
    `3115`:pep:.

    .. seealso:: `xoutil.future.types.prepare_class`:func: and
       `xoutil.future.types.new_class`:func:.

    .. versionchanged:: 1.7.1 Now are accepted atypical meta-classes, for
       example functions or any callable with the same arguments as those that
       type accepts (class name, tuple of base classes, attributes mapping).

'''

metaclass.__module__ = __name__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Extensions to the `functools` module from the Python's standard library.

You may use this module as drop-in replacement of `functools`.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_imports)

from functools import *    # noqa
from functools import _CacheInfo  # noqa

from xoutil.eight import python_version    # noqa
from xoutil.eight import callable    # noqa

import xoutil.fp.tools as fp
from xoutil.deprecation import deprecated


@deprecated(fp.pos_args)
class ctuple(fp.pos_args):
    '''Simple tuple marker for `compose`:func:.

    Since is a callable you may use it directly in ``compose`` instead of
    changing your functions to returns ctuples instead of tuples::

       >>> def compat_print(*args):
       ...     for arg in args:
       ...         print(arg)

       >>> compose(compat_print, ctuple, list, range, math=False)(3)
       0
       1
       2

       # Without ctuple prints the list
       >>> compose(compat_print, list, range, math=False)(10)
       [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    .. deprecated:: 1.8.0 Use `xoutil.fp.tool.pos_args`:class:.

    '''


# TODO: [med]  Should we port the `ctuple` to fp.tools?
@deprecated(fp.compose)
def compose(*callables, **kwargs):
    '''Returns a function that is the composition of several `callables`.

    By default `compose` behaves like mathematical function composition: this
    is to say that ``compose(f1, ... fn)`` is equivalent to ``lambda _x:
    fn(...(f1(_x))...)``.

    If any "intermediate" function returns a `ctuple`:class: it is expanded as
    several positional arguments to the next function.

    .. versionchanged:: 1.5.5 At least a callable must be passed, otherwise a
                        TypeError is raised.  If a single callable is passed
                        it is returned without change.

    :param math: Indicates if `compose` should behave like mathematical
                 function composition: last function in `funcs` is applied
                 last. If False, then the last function in `func` is applied
                 first.

    .. deprecated:: 1.8.0 Use `xoutil.fp.tools.compose`:class:.

    '''
    from xoutil.params import check_count
    check_count(callables, 1, caller='compose')
    if not all(callable(func) for func in callables):
        raise TypeError('Every func must a callable')
    if len(callables) == 1:
        return callables[0]
    math = kwargs.get('math', True)
    if not math:
        callables = list(reversed(callables))

    return fp.compose(*reversed(callables))


# TODO: Check relevance of the following function.
# The real signature should be (*funcs, times)
def power(*args):
    '''Returns the "power" composition of several functions.

    Examples::

       >>> import operator
       >>> f = power(partial(operator.mul, 3), 3)
       >>> f(23) == 3*(3*(3*23))
       True

       >>> power(operator.neg)
       Traceback (most recent call last):
       ...
       TypeError: power() takes at least 2 arguments (1 given)

    '''
    from xoutil.params import check_count
    from xoutil.fp.tools import compose
    check_count(args, 2, caller='power')
    *funcs, times = args
    if any(not callable(func) for func in funcs):
        raise TypeError('Arguments of `power`, but last, must be callables')
    if not (isinstance(times, int) and times > 0):
        raise TypeError('Last argument of `power` must be a positive integer')
    if len(funcs) > 1:
        base = (compose(funcs), )
    else:
        base = (funcs[0], )
    return compose(*(base * times))


def lwraps(*args, **kwargs):
    '''Lambda wrapper.

    Useful for decorate lambda functions with name and documentation.

    As positional arguments could be passed the function to be decorated and
    the name in any order.  So the next two ``identity`` definitions are
    equivalents::

      >>> from xoutil.future.functools import lwraps as lw

      >>> identity = lw('identity', lambda arg: arg)

      >>> identity = lw(lambda arg: arg, 'identity')

    As keyword arguments could be passed some special values, and any number
    of literal values to be assigned:

    - **name**: The name of the function (``__name__``); only valid if not
      given as positional argument.

    - **doc**: The documentation (``__doc__`` field).

    - **wrapped**: An object to extract all values not yet assigned.  These
      values are ('__module__', '__name__' and '__doc__') to be assigned, and
      '__dict__' to be updated.

    If the function to decorate is present in the positional arguments, this
    same argument function is directly returned after decorated; if not a
    decorator is returned similar to standard `wraps`:func:.

    For example::

      >>> from xoutil.future.functools import lwraps as lw

      >>> is_valid_age = lw('is-valid-human-age', lambda age: 0 < age <= 120,
      ...                   doc=('A predicate to evaluate if an age is '
      ...                        'valid for a human being.')

      >>> @lw(wrapped=is_valid_age)
      ... def is_valid_working_age(age):
      ...     return 18 < age <= 70

      >>> is_valid_age(16)
      True

      >>> is_valid_age(200)
      False

      >>> is_valid_working_age(16)
      False

    .. versionadded:: 1.7.0

    '''
    from types import FunctionType, MethodType
    from xoutil.symbols import Unset
    from xoutil.eight import string_types, iteritems
    from xoutil.eight import string
    from xoutil.params import check_count

    def repeated(name):
        msg = "lwraps got multiple values for argument '{}'"
        raise TypeError(msg.format(name))

    def settle_str(name, value):
        if value is not Unset:
            if isinstance(value, string_types):
                if name not in source:
                    source[name] = value
                else:
                    repeated(name)
            else:
                from xoutil.eight import type_name
                msg = 'lwraps expecting string for "{}", {} found'
                raise TypeError(msg.format(name, type_name(value)))

    methods = (staticmethod, classmethod, MethodType)
    decorables = methods + (FunctionType, )

    name_key = '__name__'
    doc_key = '__doc__'
    mod_key = '__module__'
    safes = {name_key, mod_key}
    source = {}
    target = Unset
    count = len(args)
    check_count(count, 0, 2, caller='lwraps')
    i = 0
    while i < count:
        arg = args[i]
        if isinstance(arg, string_types):
            settle_str(name_key, arg)
        elif isinstance(arg, decorables):
            if target is Unset:
                target = arg
            else:
                repeated('target-function')
        else:
            msg = 'lwraps arg {} must be a string or decorable function'
            raise TypeError(msg.format(i))
        i += 1
    wrapped = kwargs.pop('wrapped', Unset)
    settle_str(name_key, kwargs.pop('name', Unset))
    settle_str(name_key, kwargs.pop(name_key, Unset))
    settle_str(doc_key, kwargs.pop('doc', Unset))
    settle_str(doc_key, kwargs.pop(doc_key, Unset))
    source.update(kwargs)
    if wrapped is not Unset:
        # TODO: Check the type of `wrapped` to find these attributes in
        # disparate callable objects similarly with functions.
        for name in (mod_key, name_key, doc_key):
            if name not in source:
                source[str(name)] = getattr(wrapped, name)
        d = source.setdefault('__dict__', {})
        d.update(wrapped.__dict__)

    def wrapper(target):
        if isinstance(target, decorables):
            res = target
            if isinstance(target, methods):
                target = target.__func__
            for name in (mod_key, name_key, doc_key):
                if name in source:
                    value = source.pop(name)
                    if name in safes:
                        value = string.force(value)
                    setattr(target, str(name), value)
                d = source.pop('__dict__', Unset)
                if d:
                    target.__dict__.update(d)
            for key, value in iteritems(source):
                setattr(target, key, value)
            return res
        else:
            from xoutil.eight import type_name
            msg = 'only functions are decorated, not {}'
            raise TypeError(msg.format(type_name(target)))

    return wrapper(target) if target else wrapper

    # TODO: Next code could be removed.
    # func.__name__ = string.force(name)
    # if doc:
    #     func.__doc__ = doc
    # return func


def curry(f):
    '''Return a function that automatically 'curries' is positional arguments.

    Example::

        >>> add = curry(lambda x, y: x + y)
        >>> add(1)(2)
        3

        >>> add(1, 2)
        3

        >>> add()()()(1, 2)
        3
    '''
    from xoutil.future.inspect import getfullargspec
    fargs = getfullargspec(f)[0]

    def curried(cargs=None):
        if cargs is None:
            cargs = []

        def inner(*args, **kwargs):
            cargs_ = cargs + list(args)
            if len(cargs_) < len(fargs):
                return curried(cargs_)
            else:
                return f(*cargs_, **kwargs)
        return inner
    return curried()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Python Software Foundation
# All rights reserved.
#
# Most of the contents of this file were extracted from the source code of
# CPython 3.6.
#

'''Utilities for `with-statement contexts <343>`:pep:.

This module re-export all symbols from the standard library, with the
exception of the function `nested`.

In Python 2.7, this re-export from the package `contextlib2`_.

This module does not guarantee full compatibility with `Exception Chaining and
Embedded Tracebacks <xoutil.eight.exceptions>`:mod: and `Context Managers
<343>`:pep: back-ported in Python 2 by `contextlib2`_.

.. _contextlib2: http://contextlib2.readthedocs.io/en/stable/

.. versionadded:: 1.9.5

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import contextlib as _stdlib    # noqa
from contextlib import *    # noqa

__all__ = [n for n in getattr(_stdlib, '__all__', []) if n != 'nested']


try:
    ContextDecorator    # noqa
except NameError:
    from contextlib2 import ContextDecorator    # noqa
    __all__.append('ContextDecorator')

try:
    ExitStack    # noqa
except NameError:
    from contextlib2 import ExitStack    # noqa
    __all__.append('ExitStack')


try:
    redirect_stdout    # noqa
except NameError:
    from contextlib2 import redirect_stdout    # noqa
    __all__.append('redirect_stdout')


try:
    redirect_stderr    # noqa
except NameError:
    from contextlib2 import redirect_stderr    # noqa
    __all__.append('redirect_stderr')


try:
    suppress    # noqa
except NameError:
    from contextlib2 import suppress    # noqa
    __all__.append('suppress')


try:
    del nested    # noqa
except NameError:
    pass
